"""
Create an equity basket index
"""

from merqube_client_lib.logging import get_module_logger
from merqube_client_lib.templates.equity_baskets.multiple_equity_basket.base import (
    TYPE_SPECIFIC_OPTIONAL,
    TYPE_SPECIFIC_REQUIRED,
    get_constituents,
    inline_to_tp,
)
from merqube_client_lib.templates.equity_baskets.util import (
    create_index,
    load_template,
    read_file,
)
from merqube_client_lib.types import CreateReturn

logger = get_module_logger(__name__)


def create(config_file_path: str, prod_run: bool = False) -> CreateReturn:
    """
    Creates a new Equity Basket with multiple entries
    This class does not handle Corax.
    """
    client, template, index_info, inner_spec = load_template(
        template_name="EQUITY_BASKET_TEMPLATE_V1",
        config_file_path=config_file_path,
        type_specific_req_fields=TYPE_SPECIFIC_REQUIRED,
        type_specific_opt_fields=TYPE_SPECIFIC_OPTIONAL,
    )

    const = get_constituents(
        constituents_csv_path=index_info["constituents_csv_path"],
        base_date=index_info["base_date"],
        base_value=index_info["base_value"],
        add_initial_cash_position=True,
    )

    ports = {
        "constituents": const,
        "date_type": "EFFECTIVE",
        "identifier_type": "RIC",
        "quantity_type": "SHARES",
        "specification_type": "API",
    }

    target_portfolios = inline_to_tp(ports)

    inner_spec["portfolios"]["specification_type"] = "API"
    inner_spec["portfolios"]["constituents"] = []

    if pth := index_info.get("level_overrides_csv_path"):
        inner_spec["level_overrides"] = read_file(pth)

    post_template, ident = create_index(
        client=client,
        template=template,
        index_info=index_info,
        inner_spec=inner_spec,
        prod_run=prod_run,
        initial_target_portfolios=target_portfolios,
    )

    return CreateReturn(template=post_template, ident=ident, initial_target_ports=target_portfolios)



from typing import Any, cast

import pandas as pd

from merqube_client_lib.pydantic_types import (
    BasketPosition,
    EquityBasketPortfolio,
    PortfolioUom,
    RicEquityPosition,
)
from merqube_client_lib.templates.equity_baskets.util import read_file
from merqube_client_lib.types import TargetPortfoliosDates

TYPE_SPECIFIC_REQUIRED = ["constituents_csv_path", "base_value"]

# Optional keys that have defaults
TYPE_SPECIFIC_OPTIONAL = ["level_overrides_csv_path"]


def get_constituents(
    constituents_csv_path: str,
    base_date: str | pd.Timestamp | None = None,
    base_value: float | None = None,
    add_initial_cash_position: bool = False,
) -> list[dict[str, Any]]:
    """read and validate constituents file"""
    constituents = read_file(constituents_csv_path)

    if add_initial_cash_position:
        if not base_date or not base_value:
            raise ValueError("base_date and base_value must be set if add_initial_cash_position is True")

        if pd.Timestamp(base_date) >= pd.Timestamp(min(c["date"] for c in constituents)):
            raise ValueError("base_date should be set to a date at least one day before the first portfolio date")

        constituents += cast(
            dict[str, Any],
            [
                {
                    "date": base_date,
                    "identifier": "USD",
                    "quantity": base_value,
                    "security_type": "CASH",
                }
            ],
        )

    return constituents


def inline_to_tp(portfolio: dict[str, Any]) -> TargetPortfoliosDates:
    """
    Convert the inline portfolio to a list of target portfolios
    """

    target_portfolios: TargetPortfoliosDates = []

    uom = PortfolioUom.SHARES if portfolio["quantity_type"] == "SHARES" else PortfolioUom.WEIGHT
    id_type = portfolio["identifier_type"]

    # see how nany TP values we need
    dates = set()
    for constituent in (const := portfolio["constituents"]):
        dates.add(pd.Timestamp(constituent["date"]))

    for date in sorted(dates):
        positions: list[BasketPosition | RicEquityPosition] = []
        for constituent in const:
            sec_type = constituent["security_type"]
            pos_class = RicEquityPosition if sec_type == "EQUITY" else BasketPosition

            if pd.Timestamp(constituent["date"]) != date:
                continue  # will get picked up in another target portfolio

            positions.append(
                pos_class(
                    asset_type=constituent["security_type"],
                    identifier=constituent["identifier"],
                    amount=constituent["quantity"],
                    # this is only at the top level of inline; but its wrong for the cash position (in TP it is CURRENCY_CODE)
                    identifier_type="CURRENCY_CODE" if sec_type == "CASH" else id_type,
                )
            )

        target_port_val = EquityBasketPortfolio(timestamp=date, unit_of_measure=uom, positions=positions)

        EquityBasketPortfolio.parse_obj(target_port_val)  # validate
        target_portfolios.append((date, target_port_val))

    return target_portfolios



"""
Replace the portfolio of an equity basket index (or post future portfolios)
"""
import os

import click
import pandas as pd

from merqube_client_lib.api_client.merqube_client import MerqubeAPIClientSingleIndex
from merqube_client_lib.logging import get_module_logger
from merqube_client_lib.templates.equity_baskets.multiple_equity_basket.base import (
    get_constituents,
    inline_to_tp,
)
from merqube_client_lib.templates.equity_baskets.util import get_inner_spec, read_file
from merqube_client_lib.util import get_token, pydantic_to_dict

logger = get_module_logger(__name__)


def _update_portfolio(
    index_name: str, constituents_csv_path: str, level_overrides_csv_path: str | None = None, prod_run: bool = False
) -> None:
    """
    Creates a new Equity Basket with multiple entries
    """
    client = MerqubeAPIClientSingleIndex(index_name=index_name, token=get_token())
    template = client.get_model()

    ports = {
        "constituents": get_constituents(constituents_csv_path=constituents_csv_path),
        "date_type": "EFFECTIVE",
        "identifier_type": "RIC",
        "quantity_type": "SHARES",
        "specification_type": "API",
    }

    target_portfolios = inline_to_tp(ports)

    if level_overrides_csv_path:
        inner_spec = get_inner_spec(template=template)
        inner_spec["level_overrides"] = read_file(level_overrides_csv_path)

    if not prod_run:
        logger.debug("Dry run, not updating portfolio")
        return

    logger.info("Updating portfolio")
    # in the ongoing update case, we only care about dates >= today since all history is
    # already in on the server, and not relevent to the next run of the index
    today = pd.Timestamp.today()
    for date, tp in target_portfolios:
        if date >= today:
            # see the docstring under replace_target_portfolio; this works for future dates too, desptie 'replace'
            client.replace_portfolio(target_portfolio=pydantic_to_dict(tp))


@click.command()
@click.option("--index-name", type=str, required=True, help="the index name to update")
@click.option("--constituents-csv-path", type=str, required=True, help="path to the constituents csv file")
@click.option(
    "--level_overrides_csv_path",
    type=str,
    required=False,
    help="path to a level overrides csv file, if applicable (not required)",
)
@click.option("--prod-run", is_flag=True, default=False, help="Create the index in production")
def main(
    index_name: str, constituents_csv_path: str, level_overrides_csv_path: str | None = None, prod_run: bool = False
) -> None:
    """main entrypoint"""
    assert os.path.exists(constituents_csv_path), f"Constituents csv file does not exist: {constituents_csv_path}"
    if level_overrides_csv_path:
        assert os.path.exists(
            level_overrides_csv_path
        ), f"Level overrides csv file does not exist: {level_overrides_csv_path}"
    _update_portfolio(index_name=index_name, constituents_csv_path=constituents_csv_path, prod_run=prod_run)


if __name__ == "__main__":
    main()  # pyright: ignore  # pylint: disable=no-value-for-parameter



"""
Merqube custom types
"""
from dataclasses import dataclass
from enum import Enum, unique
from typing import Any, Final

import pandas as pd

from merqube_client_lib.pydantic_types import (
    EquityBasketPortfolio,
    IdentifierUUIDPost,
    IndexDefinitionPost,
)

HTTP_METHODS: Final[list[str]] = ["GET", "DELETE", "POST", "GET", "PUT", "OPTIONS", "PATCH"]


@unique
class HTTPMethod(str, Enum):
    GET = "GET"
    PUT = "PUT"
    DELETE = "DELETE"
    PATCH = "PATCH"
    OPTIONS = "OPTIONS"
    POST = "POST"
    HEAD = "HEAD"


Manifest = dict[str, Any]
ManifestList = list[Manifest]

Records = list[dict[str, str]]

TargetPortfoliosDates = list[tuple[pd.Timestamp, EquityBasketPortfolio]]


@dataclass
class CreateReturn:
    template: IndexDefinitionPost
    ident: IdentifierUUIDPost | None = None
    initial_target_ports: TargetPortfoliosDates | None = None



"""
Combined API Client for all merqube APIs
"""
from typing import Any, cast

import pandas as pd
from cachetools import LRUCache, cached

from merqube_client_lib.api_client import base
from merqube_client_lib.pydantic_types import EquityBasketPortfolio, IdentifierUUIDPost
from merqube_client_lib.pydantic_types import IndexDefinitionPatchPutGet as Index
from merqube_client_lib.pydantic_types import Provider
from merqube_client_lib.session import MerqubeAPISession
from merqube_client_lib.types import Manifest
from merqube_client_lib.util import pydantic_to_dict


class MerqubeAPIClient(base._IndexAPIClient, base._SecAPIClient):  # noqa
    """
    Combined API Client for indexapi + secapi, for methods that do not pertain to a particular index
    (general queries, CRUD of indices, etc)
    For a client pertaining to a specific existing index, use MerqubeAPIClientSingleIndex
    """

    def create_identifier(
        self, provider: Provider, identifier_post: IdentifierUUIDPost | dict[str, Any]
    ) -> dict[str, Any]:
        """
        Create a new identifier (in MerQubes system - this does NOT talk to the provider)
        """
        name = (
            payload := (
                pydantic_to_dict(identifier_post)
                if isinstance(identifier_post, IdentifierUUIDPost)
                else identifier_post
            )
        )["name"]
        linked_name = payload["index_name"]

        results = self.session.get_collection(f"/identifier/{provider.value}?names={name}")
        if len(results) == 0:
            return cast(dict[str, Any], self.session.post(f"/identifier/{provider.value}", json=payload).json())

        if (exis_name := results[0]["index_name"]) == linked_name:
            return {"status": "already exists for this index name"}
        raise ValueError(f"Identifier {name} already exists for a different index name ({exis_name})")


class MerqubeAPIClientSingleIndex(MerqubeAPIClient):
    """
    Class that contains methods that deal with a single index
    Instantiate with an index name ("name" field in the index manifest) and indicate if it is a realtime index
    """

    def __init__(self, index_name: str, is_intraday: bool = False, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        res = self.session.get_collection_single(f"/index?name={index_name}")
        self.index_id = res["id"]
        self.is_intraday = is_intraday
        self.index_name = index_name

        self.mod = mod = self.get_index_model(index_name=index_name)

        # intraday model has a nasty type of [None | Intraday | Bool ..]
        self._has_intraday = mod.intraday is not None and (
            (isinstance(mod.intraday, bool) and mod.intraday) or mod.intraday.enabled is True
        )

        # get the security id for this index
        self._sec_id = self.session.get_collection_single(f"/security/index?name={index_name}")["id"]
        self._intra_sec_id = (
            self.session.get_collection_single(f"/security/intraday_index?name={index_name}")["id"]
            if self._has_intraday
            else None
        )

    def get_manifest(self) -> Manifest:
        """
        Get index definition for specified index id
        """
        return self.get_index_manifest(index_name=self.index_name)

    def get_model(self) -> Index:
        """
        Get index model for specified index id
        """
        return self.mod

    def get_metrics(
        self,
        metrics: list[str],
        use_intraday_metrics: bool = False,
        start_date: pd.Timestamp | None = None,
        end_date: pd.Timestamp | None = None,
    ) -> pd.DataFrame:
        """
        Get a list of metrics for this index
        Some indices are both end of day and intraday - use the flag to query for intraday metrics
        """
        return self.get_security_metrics(
            sec_type="intraday_index" if use_intraday_metrics else "index",
            sec_ids=[cast(str, self._intra_sec_id) if use_intraday_metrics else self._sec_id],
            metrics=metrics,
            start_date=start_date,
            end_date=end_date,
        )

    def get_returns(
        self,
        returns_metric: str = "price_return",
        use_intraday_metrics: bool = False,
        start_date: pd.Timestamp | None = None,
        end_date: pd.Timestamp | None = None,
    ) -> pd.DataFrame:
        """
        Get returns for this index
        Some indices are both end of day and intraday - use the flag to query for intraday metrics
        """
        if use_intraday_metrics and not self._has_intraday:
            raise ValueError("This index is not an intraday index")

        return self.get_metrics(
            metrics=[returns_metric],
            use_intraday_metrics=use_intraday_metrics,
            start_date=start_date,
            end_date=end_date,
        )

    def get_portfolio(self) -> list[dict[str, Any]]:
        """
        Get list of portfolios for specified index id
        Each entry in the array has a date, and it's the portfolio that was active on that (rebalance) date
        """
        return self.session.get_collection(f"/index/{self.index_id}/portfolio")

    def get_portfolio_allocations(self) -> list[dict[str, Any]]:
        """
        Get list of portfolio allocations for specified index id
        This is a view of how the portfolio has changed over time
        """
        return self.session.get_collection(f"/index/{self.index_id}/portfolio_allocations")

    def get_target_portfolio(
        self, start_date: pd.Timestamp | None = None, end_date: pd.Timestamp | None = None
    ) -> list[dict[str, Any]]:
        """
        Some indices allow the customer to set a target portfolio (e.g., equity baskets) and a target date
        this returns those target_portfolios
        sometime after the date of each, a get on /portfolio may return the target portfolio (but in some cases it may not be possible to achieve the exact target)
        """
        opts = {}
        if start_date is not None:
            opts["start_date"] = start_date.date().isoformat()

        if end_date is not None:
            opts["end_date"] = end_date.date().isoformat()

        return self.session.get_collection(f"/index/{self.index_id}/target_portfolio", options=opts)

    def get_caps(self) -> list[dict[str, Any]]:
        """
        Get list of caps for specified index id (only applies to buffer indices)
        """
        return self.session.get_collection(f"/index/{self.index_id}/caps")

    def get_stats(self) -> list[dict[str, Any]]:
        """
        Get stats, which are historical returns over different time periods, of the index
        """
        return self.session.get_collection(f"/index/{self.index_id}/stats")

    def get_data_collections(self) -> list[dict[str, Any]]:
        """
        Get list of data collections for specified index id

        Data Collections are a system for retrieving data on a daily basis after calculations have been performed.  After data has been disseminated for public usage on various providers like Reuters, Bloomberg, etc, clients may opt-in to receive this data directly from MerQube.  We may also provide intermediate data that is not directly disseminated which might be useful for clients.

        This is not available for all indices
        """
        return self.session.get_collection(f"/index/{self.index_id}/data_collections")

    def replace_portfolio(self, target_portfolio: dict[str, Any] | EquityBasketPortfolio) -> dict[str, Any]:
        """
        set the open portfolio for a given day. see the base class docs for more info
        """
        tp = (
            pydantic_to_dict(target_portfolio)
            if isinstance(target_portfolio, EquityBasketPortfolio)
            else target_portfolio
        )
        return self.replace_target_portfolio(index_id=self.index_id, target_portfolio=tp)


client_cache: LRUCache = LRUCache(maxsize=256)  # type: ignore


@cached(cache=client_cache)
def get_client(
    index_name: str | None = None,
    is_intraday: bool = False,
    user_session: MerqubeAPISession | None = None,
    token: str | None = None,
    **session_kwargs: Any,
) -> MerqubeAPIClient | MerqubeAPIClientSingleIndex:
    """
    Cached; returns a Merqube client (or SingleClient if provided an index name) for token
    """
    return (
        MerqubeAPIClientSingleIndex(
            index_name=index_name, is_intraday=is_intraday, user_session=user_session, token=token, **session_kwargs
        )
        if index_name
        else MerqubeAPIClient(user_session=user_session, token=token, **session_kwargs)
    )



"""
Misc helper utilities
"""
import datetime
import json
import os
from typing import Any, cast

import pandas as pd


def batch_post_payload(rows: list[Any], batch_size: int) -> list[list[Any]]:
    """
    returns a list of sublists of size <= batch_size
    used widely for POSTing large payloads to e.g., secapi
    If you have veryy large payloads, you may want to consider doing this math virtually rather than recopying.
    """
    # If rows is `None` or is an empty list, it doesn't need to be a list of sublist at all
    if rows is None or len(rows) == 0:
        return []
    if len(rows) > batch_size:
        n = batch_size
        return [rows[i * n : (i + 1) * n] for i in range((len(rows) + n - 1) // n)]
    return [rows]


def freezable_now(tz: str = "UTC", no_tzinfo_iso: bool = False) -> pd.Timestamp | str:
    """
    Builds pd.Timestamp.now out of a datetime.datetime so it can be frozen in unit tests
    """
    dt = datetime.datetime.utcnow()
    ts = pd.Timestamp(dt, tz="UTC").tz_convert(tz)
    if not no_tzinfo_iso:
        return cast(pd.Timestamp, ts)
    return ts.tz_localize(None).isoformat()  # pyright: ignore


def freezable_utcnow(no_tzinfo_iso: bool = False) -> pd.Timestamp | str:
    """
    Builds pd.Timestamp.utcnow out of a datetime.datetime so it can be frozen in unit tests
    """
    return freezable_now(no_tzinfo_iso=no_tzinfo_iso)


def get_token() -> str:
    """
    Gets the users token from the env var
    """
    if not (api_key := os.environ.get("MERQ_API_KEY")):
        raise ValueError("MERQ_API_KEY not set")
    return api_key


def pydantic_to_dict(pydantic_obj: Any) -> dict[str, Any]:
    """
    helper function to convert pydantic objects to dictionaries
    you would think .dict() would work - but that can still contain objects such as Enums,
    which are not json serializable
    """
    return cast(dict[str, Any], json.loads(pydantic_obj.json(exclude_none=True, exclude_defaults=True)))



"""
Module for mocking secapi, mostly for unit testing, but also for other components testing against secapi
"""
import importlib
import logging
from typing import Any, Callable, Optional
from unittest.mock import MagicMock

from cachetools import Cache

from merqube_client_lib.logging import get_module_logger

logger = get_module_logger(__name__, level=logging.DEBUG)


def _sec_types() -> list[dict[str, str]]:
    """fixed map for mocking valid sectypes"""
    return [
        {"name": "crypto_asset"},
        {"name": "custom"},
        {"name": "equity"},
        {"name": "exchange"},
        {"name": "futures_contract"},
        {"name": "futures_option"},
        {"name": "futures_root"},
        {"name": "fx"},
        {"name": "index"},
        {"name": "intraday_index"},
        {"name": "interest_rate"},
        {"name": "economic_data"},
    ]


def mock_secapi_builder(
    get_session_function_path: str, get_client_function_path: str, secapi_client_cache: Optional[Cache] = None  # type: ignore
) -> Callable[..., None]:
    """
    Builds the `mock_secapi` function used for the API for its tests.

    ```
    mock_secapi = mock_secapi_builder(
        get_session_function_path="dataapi.runners.sec_client._get_dataapi_session",
        get_client_function_path="dataapi.runners.sec_client.get_client",
        secapi_client_cache=sec_client.client_cache
    )
    ```

    Then in your tests, you can use `mock_secapi`

    ```
    mock_secapi(
        monkeypatch=monkeypatch,
        method_name_function_map={
            "get_security_definitions_mapping_table": MagicMock(return_value={}),
            "_post_dedupe": MagicMock(),
            ...
        },
    )
    ```
    """
    for function in [get_session_function_path, get_client_function_path]:
        # Checks if all the function exists
        importlib.find_loader(function)

    client_function_module_name = get_client_function_path[: get_client_function_path.rindex(".")]
    client_module = importlib.import_module(client_function_module_name)

    def mock_secapi(
        monkeypatch,
        method_name_function_map: dict[str, Callable[..., Any]],
        session_func_map: dict[str, Callable[..., Any]] | None = None,
        **get_client_kwargs: Any,
    ) -> None:
        """
        get a client with some methods patched with swapout functions
        """
        if secapi_client_cache is not None:
            logger.debug("Clearing client cache")
            secapi_client_cache.clear()
        else:
            logger.debug("No client cache to clear")

        sess = MagicMock()

        if session_func_map is not None:
            for method_name, func in session_func_map.items():
                if getattr(sess, method_name) is None:
                    raise ValueError(f"Trying to patch a nonexisting session method: {method_name}")
                logger.debug(f"Mocking session method {method_name}")
                setattr(sess, method_name, func)

        def get_session(*args: Any, **kwargs: Any):
            logger.debug("Returning mocked session")
            return sess

        logger.debug(f"Mocking get_session call at {get_session_function_path}")
        monkeypatch.setattr(get_session_function_path, get_session)

        client = client_module.get_client(**get_client_kwargs)

        if "get_supported_secapi_types" not in method_name_function_map:
            setattr(client, "get_supported_secapi_types", _sec_types)

        for method_name, func in method_name_function_map.items():
            if getattr(client, method_name) is None:
                raise ValueError(f"Trying to patch a nonexisting client method: {method_name}")
            logger.debug(f"Mocking method {method_name}")
            setattr(client, method_name, func)

        def get_client(*args: Any, **kwargs: Any):
            logger.debug("Returning mocked client")
            return client

        # now a get_client call from the main code will return this mocked client:
        logger.debug(f"Mocking get_client call at {get_client_function_path}")
        monkeypatch.setattr(get_client_function_path, get_client)

    return mock_secapi



"""
Create an equity basket index
"""
import logging
import os
import tempfile
from typing import cast

import boto3
import pandas as pd

from merqube_client_lib.exceptions import APIError
from merqube_client_lib.logging import get_module_logger
from merqube_client_lib.templates.equity_baskets.util import create_index, load_template
from merqube_client_lib.types import CreateReturn, Records

logger = get_module_logger(__name__, level=logging.DEBUG)

TYPE_SPECIFIC_REQUIRED = [
    "underlying_index_name",
    "fee_value",
    "fee_type",
    "day_count_convention",
    "base_value",
    "client_owned_underlying",
]

TYPE_SPECIFIC_OPTIONAL: list[str] = []


def _download_tr_map(tdir: str) -> str:
    """broken out for ease of unit testing"""
    s3 = boto3.client("s3")
    s3.download_file("merq-api-assets", "tr_indices.csv", (pth := os.path.join(tdir, "tr.csv")))
    return pth


def _get_trs(tr: str | None = None) -> Records:
    """
    Returns the list of all MerQube TRs decrements can be on
    """
    with tempfile.TemporaryDirectory() as tdir:
        trs = pd.read_csv(_download_tr_map(tdir))

    if trs.empty:
        raise ValueError("No TRs found")

    trs = trs.sort_values(by=["underlying_ric"])

    if not tr:
        return cast(Records, trs.to_dict(orient="records"))

    row = trs[trs["index_name"] == tr]
    if row.empty:
        raise ValueError(f"{tr} is not a valid TR")

    return cast(Records, row.to_dict(orient="records"))


def create(config_file_path: str, prod_run: bool = False) -> CreateReturn:
    """
    Creates a new Equity Basket with multiple entries
    This class does not handle Corax.
    """
    client, template, index_info, inner_spec = load_template(
        template_name="DECREMENT_TEMPLATE_VERSION_1",
        config_file_path=config_file_path,
        type_specific_req_fields=TYPE_SPECIFIC_REQUIRED,
        type_specific_opt_fields=TYPE_SPECIFIC_OPTIONAL,
    )

    uin = index_info["underlying_index_name"]
    if index_info["client_owned_underlying"]:
        # check that the underlying index exists
        logger.info(f"Checking that the underlying index {uin} exists")
        try:
            client.get_index_model(index_name=uin)
        except APIError:
            raise ValueError(f"{uin} is not a valid index or you are not permissioned for it")

        underlying_secapi_metric = "price_return"
    else:
        logger.info(f"Checking that the underlying index {uin} is a permissioned MerQube provided TR")
        tr = _get_trs(uin)[0]
        underlying_secapi_metric = tr["metric"]

    # set the metric to the TR metric
    template.plot_metric = inner_spec.get("output_metric", (getattr(template, "plot_metric", "price_return")))

    inner_spec["metric"] = underlying_secapi_metric  # the decrement index looks up the returns of the tr via this
    inner_spec["underlying_ticker"] = uin

    if (
        not isinstance(fv := index_info["fee_value"], (int, float))
        or not isinstance(ft := index_info["fee_type"], str)
        or ft not in ["fixed", "percentage_pre", "percentage_post"]
    ):
        raise ValueError(
            "fee_value must be a number and fee_type must be one of fixed, percentage_pre, percentage_post"
        )

    inner_spec["day_count_convention"] = index_info["day_count_convention"]

    inner_spec["fee"] = {"fee_value": fv, "fee_type": ft}

    post_template, ident_ppost = create_index(
        client=client, template=template, index_info=index_info, inner_spec=inner_spec, prod_run=prod_run
    )

    return CreateReturn(post_template, ident_ppost)



"""
Module that handles logging
"""
import logging
import logging.handlers
import os

from merqube_client_lib.constants import (
    MERQ_REQUEST_ID_ENV_VAR,
    MERQ_REQUEST_REMOTE_ADDR_ENV_VAR,
    MERQ_RUN_ID_ENV_VAR,
)


def _get_file_handler(filename: str) -> logging.Handler:
    # 100MB log, up to 5 (enhancement: take these as params)
    handler = logging.handlers.RotatingFileHandler(filename, maxBytes=100000000, backupCount=5)
    handler.setFormatter(FORMATTER)

    return handler


class MerqFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        log_str = super().format(record)
        req_id = os.getenv(MERQ_REQUEST_ID_ENV_VAR)
        run_id = os.getenv(MERQ_RUN_ID_ENV_VAR)
        remote_addr = os.getenv(MERQ_REQUEST_REMOTE_ADDR_ENV_VAR)
        if req_id is None and run_id is None and remote_addr is None:
            return log_str

        splice = ""
        if req_id is not None:
            splice += f" [req-id:{req_id}]"
        if run_id is not None:
            splice += f" [run-id:{run_id}]"
        if remote_addr is not None:
            splice += f" [ip:{remote_addr}]"

        # Move the request id after time
        time_offset = 25
        return log_str[0:time_offset] + splice + log_str[time_offset:]


FORMATTER = MerqFormatter("[%(asctime)s] [%(name)s] [%(levelname)s] [%(process)d] %(message)s")


def get_module_logger(
    mod_name: str, level: int = logging.DEBUG, filename: str | None = None, propagate: bool = False
) -> logging.Logger:
    """
    Call this from each module with __name__
    see https://docs.python.org/3/howto/logging.html#advanced-logging-tutorial
    """
    logger = logging.getLogger(mod_name)
    logger.setLevel(level)

    if not propagate:
        logger.propagate = False

    handler = logging.StreamHandler()
    handler.setFormatter(FORMATTER)
    logger.addHandler(handler)

    if filename:
        fhandler = _get_file_handler(filename)
        logger.addHandler(fhandler)

    return logger
